theory 
  Scenario4
imports 
  FWTesting
begin

subsubsection{* Network Model *}
text{* We now have four subnetworks instead of only two as before. *}

definition net1 :: "int set" where "net1 = {x. x  > 1 \<and> x < 100}"
definition net2 :: "int set" where "net2 = {x. x  > 1001 \<and> x < 1056}"
definition net3 :: "int set" where "net3 = {x. x  > 2001 \<and> x < 2056}"
definition net4 :: "int set" where "net4 = {x. x  > 3001 \<and> x < 3056}"


definition
 subnet1 :: "IntegerPort_TCPUDP net" where
"subnet1 = {{(a,b,c). a \<in> net1}}"

definition
 subnet2 :: "IntegerPort_TCPUDP net" where
"subnet2 = {{(a,b,c). a \<in> net2}}"

definition
 subnet3 :: "IntegerPort_TCPUDP net" where
"subnet3 = {{(a,b,c). a \<in> net3}}"

definition
 subnet4 :: "IntegerPort_TCPUDP net" where
"subnet4 = {{(a,b,c). a \<in> net4}}"

subsubsection{* Policy Model *}                    

text{* If we are using policy normalisation, we need the syntactic
policy rules. There is only a small fixed number of them and it is
important to note that they are processed from left to right. *}

definition Policy :: "(IntegerPort_TCPUDP net, protocol \<times> port) Combinators" 
 where 
 "Policy = (DenyAll \<oplus> 
AllowPortFromTo subnet1 subnet2 (tcp, (80::port)) \<oplus> 
AllowPortFromTo subnet1 subnet3 (tcp,(8080::port)) \<oplus> 
AllowPortFromTo subnet4 subnet3 (udp, (54::port)) \<oplus>
AllowPortFromTo subnet3 subnet2 (tcp, (32::port)) \<oplus> 
AllowPortFromTo subnet1 subnet4 (udp, (65::port)) \<oplus>
AllowPortFromTo subnet2 subnet4 (tcp, (861::port)))"

subsubsection{* Definitions of the testing constraints. *} 
text{* We use the standard testing constraints: *}

definition
 accross_subnets :: "(IntegerPort_TCPUDP,'b) packet \<Rightarrow> bool" where
"accross_subnets x = (((src x \<sqsubset> subnet1) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet1))) \<and>  
                    ((src x \<sqsubset> subnet2) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet2))) \<and>  
                    ((src x \<sqsubset> subnet3) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet3))) \<and>  
                    ((src x \<sqsubset> subnet4) \<longrightarrow> (\<not> (dest x \<sqsubset> subnet4))) \<and>
    (src x \<sqsubset> subnet1 \<or> src x \<sqsubset> subnet2 \<or> src x \<sqsubset> subnet3 \<or> src x \<sqsubset> subnet4) \<and>  
    (dest x \<sqsubset> subnet1 \<or> dest x \<sqsubset> subnet2 \<or> dest x \<sqsubset> subnet3 \<or> dest x \<sqsubset> subnet4))"



subsubsection{* Test Generation *}

text{* The additional lemmas required now are Cp.simps and dom\_def/ *}

lemmas policyLemmas = ProtocolPortCombinators.ProtocolCombinators 
                     IntegerPort_TCPUDPLemmas IntegerPort_TCPUDPTestConstraints
                  packet_defs Cp.simps dom_def Policy_def 
                  accross_subnets_def subnet1_def subnet2_def
                  subnet3_def subnet4_def                   

lemmas nets = net1_def net2_def net3_def net4_def

text{* We need to perform some proof work needed for the
normalisation. We start with stating that there is a member in each
subnetwork, followed by the fact that all subnetworks are distinct,
that no policy rule has an empty domain, that all rules are different,
a technical lemma for the normalisation, a theorem that the policy is
total, and the proof that the normalisation keeps the semantics of the
policy.  *}


lemma ex1: "(\<exists>a. a \<in> net1)"
apply (auto simp: nets)
apply (rule_tac x= 13 in exI)
by auto


lemma ex2: "(\<exists>a. a \<in> net2)"
apply (auto simp: nets)
apply (rule_tac x= 1003 in exI)
by auto


lemma ex3: "(\<exists>a. a \<in> net3)"
apply (auto simp: nets)
apply (rule_tac x= 2013 in exI)
by auto

lemma ex4: "(\<exists>a. a \<in> net4)"
apply (auto simp: nets)
apply (rule_tac x= 3013 in exI)
by auto


lemma nets_different: "subnet1 \<noteq> subnet2 \<and> subnet1 \<noteq> subnet3 \<and> 
                       subnet1 \<noteq> subnet4 \<and> subnet2 \<noteq> subnet4 \<and>
                       subnet2 \<noteq> subnet1 \<and> subnet2 \<noteq> subnet3 \<and>  
                       subnet3 \<noteq> subnet1 \<and> subnet3 \<noteq> subnet2 \<and> 
                       subnet3 \<noteq> subnet4 \<and> subnet4 \<noteq> subnet1 \<and> 
                       subnet4 \<noteq> subnet2 \<and> subnet4 \<noteq> subnet3"
apply (simp add: policyLemmas )
apply (rule conjI, rule setDistinct1, simp add: nets setDistinct4,  
  smt CollectD  net2_def ex1 ex2 ex3 ex4 net1_def net3_def net4_def)+
apply (rule setDistinct1, simp add: nets setDistinct4, 
   smt CollectD ex2 net2_def ex1 ex3 ex4 net1_def net3_def net4_def)+
done

lemma noMT: "\<forall> x \<in> set (policy2list Policy). dom (Cp x) \<noteq> {}" 
apply (simp add: policyLemmas ex1 ex2 ex3 ex4)
done


lemma ANDex: "allNetsDistinct (policy2list Policy)"
apply (simp add: Policy_def allNetsDistinct_def nets_different)
apply (auto simp: policyLemmas nets netsDistinct_def)
done


lemma all_in_list: "all_in_list (policy2list Policy) (Nets_List Policy)"
apply (simp add: Policy_def)
apply (unfold Nets_List_def)
apply (unfold bothNets_def)
apply (insert nets_different,simp)
done
 

lemma PolicyUniv: "dom (Cp Policy) = UNIV" 
apply (simp add: Policy_def)
apply (rule domDA1Cp)
done



lemma C_Policy_compile: " 
Cp (list2policy (normalizePr Policy)) =  Cp Policy"
apply (unfold normalizePr_def)
apply (rule C_eq_compile)
apply (simp add: Policy_def)
apply (rule all_in_list)
apply (rule ANDex)
done


text{* The test specification is the same with the only exception that we need to state \texttt{Cp Policy} instead of only \texttt{Policy}. 
There is however more proof work required before applying test case generation. However, this is again standard for all policies of this kind. 
*}

lemma net_noteq12: "(x \<in> net1) \<Longrightarrow> (x \<notin> net2)" 
by (auto simp: nets)
lemma net_noteq13: "(x \<in> net1) \<Longrightarrow> (x \<notin> net3)" 
by (auto simp: nets)
lemma net_noteq14: "(x \<in> net1) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)
lemma net_noteq23: "(x \<in> net2) \<Longrightarrow> (x \<notin> net3)" 
by (auto simp: nets)
lemma net_noteq24: "(x \<in> net2) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)
lemma net_noteq34: "(x \<in> net3) \<Longrightarrow> (x \<notin> net4)" 
by (auto simp: nets)

lemmas net_noteq = net_noteq12 net_noteq13 net_noteq14
                   net_noteq23 net_noteq24 net_noteq34



test_spec 
 "port_positive x \<and> fix_values x \<and>
  accross_subnets x \<longrightarrow> 
 FUT x = (Cp Policy) x"
apply (prepare_fw_spec)
apply (rule_tac PL = "((Comb2PolList (rev (normalizePr (Policy))) Cp))" 
   in distrFUTLG1)
apply (simp add: PolicyUniv)
apply (rule fun_cong) back
apply (rule p_eq2)
apply (rule norm_notMT)
apply (simp add: Policy_def)
apply (rule C_Policy_compile)
apply (insert nets_different noMT)
apply (simp add:  normalizePr_def Policy_def bothNets_def aux aux2 
                  Nets_List_def)
apply (rotate_tac -1, thin_tac "?X",rotate_tac -1, thin_tac "?X")
apply (simp add: FUTListGD_def)
apply (simp add: gatherDomain_def)
apply (simp add: domain_reasoningP dom_2 domDA1)
apply (tactic  
 "REPEAT(CHANGED(TRYALL((rtac @{thm conjI}) THEN' (rtac @{thm impI}))))")
apply (simp_all add: policyLemmas)
apply (simp_all split:option.splits decision.splits prod.splits add: deny_all_def)
apply (gen_test_cases FUT simp: net_noteq)
apply simp_all
apply (tactic 
 "ALLGOALS(TestGen.ONLY_POS(
      (full_simp_tac (simpset_of @{context}addsimps @{thms policyLemmas PO_def nets}))
      THEN_ALL_NEW ( (rtac @{thm PO_add}))))")
store_test_thm "Scenario4"

declare[[testgen_SMT=true]]
declare [[testgen_iterations=15]]
 
gen_test_data "Scenario4" 


text{* There are 328 test cases being generated. While test case
generation is fairly quick (ca. 15 seconds), test data generation takes
much longer. The latter is a characteristic if using big Integer set
networks.*}

export_fw_data "Scenario4.data" Scenario4

end
