theory FTP_With_Policy
imports FWTesting
begin


definition 
  intranet :: "IntegerPort net" where
  "intranet = {{(a,e) . a = 3}}"

definition 
  internet :: "IntegerPort net" where
  "internet = {{(a,c). a > 3}}"

definition 
  ftp_policy :: "(IntegerPort,ftp_msg) FWPolicy" where
  "ftp_policy = deny_all ++ allow_from_to_port (21::port) intranet  internet"


text{* The next two constants check if an address is in the Intranet
or in the Internet respectively.*}

definition 
  is_in_intranet :: "IntegerPort  \<Rightarrow> bool" where
  "is_in_intranet a = ((fst a) = 3)"

definition 
  is_in_internet :: "IntegerPort \<Rightarrow> bool" where
  "is_in_internet a =  ((fst a) >3)"


text{*
The next definition is our starting state: an empty trace and the just defined policy.*}

definition 
  "\<sigma>_0_ftp" ::  "(id\<rightharpoonup>port) \<times>
                 (IntegerPort, ftp_msg) FWPolicy" 
where                                     
  "\<sigma>_0_ftp = (\<emptyset>,ftp_policy)"


(*
definition 
  NB_ftp :: "IntegerPort src \<Rightarrow> IntegerPort dest \<Rightarrow> id \<Rightarrow> port  \<Rightarrow>
            (IntegerPort,ftp_msg) history set" where
  "NB_ftp s d i p = {x. (is_ftp S0 s d i p x)}"

*)
definition "accept_ftp" ::  "(IntegerPort, ftp_msg) packet list \<Rightarrow> bool" where
          "accept_ftp t = 
  (\<exists> c s i p. t \<in> NB_ftp c s i p \<and> is_in_intranet c \<and> is_in_internet s \<and> (snd s) = 21)"



fun packet_with_id where
 "packet_with_id [] i = []"
|"packet_with_id (x#xs) i = (if id x = i then (x#(packet_with_id xs i)) else (packet_with_id xs i))"


text{*The depth of the test case generation corresponds to the maximal
length of generated traces. 4 is the minimum to get a full FTP
protocol run.  *}



declare [[testgen_depth=4]]
declare [[testgen_profiling]]




fun ids1 where
 "ids1 i (x#xs) = (id x = i \<and> ids1 i xs)"
|"ids1 i [] = True"


lemmas STSimps =  id_def p_accept_def
 Let_def valid_SE_def orelse_def unit_SE_def bind_SE_def
 in_subnet_def src_def dest_def IntegerPort.dest_port_def
 subnet_of_def id_def port_open_def is_init_def is_data_def
 is_port_request_def is_close_def p_accept_def content_def
 PolicyCombinators PortCombinators is_in_intranet_def
 is_in_internet_def intranet_def internet_def exI subnetOf_lemma
 subnetOf_lemma2 subnetOf_lemma3 subnetOf_lemma4 port_open_lemma
 ftp_policy_def allow_from_to_port_def
 deny_from_to_port_def
p_accept_def port_open_def subnet_of_def deny_all_def map_add_def allow_all_def
 in_subnet_def IntegerPort.subnet_of_def IntegerPort.dest_port_def

 
lemma POL1: "( (\<lambda>x. option_case (deny_all x) Some
                 (if snd (fst (snd (snd x))) = 21
                  then if (case fst (snd x) of (a, e) \<Rightarrow> a = 3) \<and>
                          (case fst (snd (snd x)) of (a, c) \<Rightarrow> 3 < a)
                       then allow_all x else None
                  else None))) =  (ftp_policy)"
apply (simp add: STSimps)
done

 
lemma POL2: "(\<lambda>x. option_case
                 (option_case (deny_all x) Some
                   (if snd (fst (snd (snd x))) = 21
                    then if (case fst (snd x) of (a, e) \<Rightarrow> a = 3) \<and>
                            (case fst (snd (snd x)) of (a, c) \<Rightarrow> 3 < a)
                         then allow_all x else None
                    else None))
                 Some
                 (if snd (fst (snd (snd x))) = X
                  then if (case fst (snd x) of (a, b) \<Rightarrow> a = p) \<and>
                          (case fst (snd (snd x)) of (a, b) \<Rightarrow> a = 3)
   then allow_all x else None else None)) = 
    allow_from_to_port (X::port) (subnet_of (p,3)) intranet \<Oplus> ftp_policy"
apply (rule ext)
apply (simp add: STSimps)
done



lemma POL3: "(\<lambda>x. option_case
                 (option_case
                   (option_case (deny_all x) Some
                     (if snd (fst (snd (snd x))) = 21
                      then if (case fst (snd x) of (a, e) \<Rightarrow> a = 3) \<and>
                              (case fst (snd (snd x)) of (a, c) \<Rightarrow> 3 < a)
                           then allow_all x else None
                      else None))
                   Some
                   (if snd (fst (snd (snd x))) = X
                    then if (case fst (snd x) of (a, b) \<Rightarrow> a = X26X3) \<and>
                            (case fst (snd (snd x)) of (a, b) \<Rightarrow> a = 3)
                         then allow_all x else None
                    else None))
                 Some
                 (if snd (fst (snd (snd x))) = X
                  then if (case fst (snd x) of (a, b) \<Rightarrow> a = X26X3) \<and>
                          (case fst (snd (snd x)) of (a, b) \<Rightarrow> a = 3)
                       then deny_all x else None
                  else None))
= ftp_policy"
apply (rule ext)
apply (simp add: STSimps)
apply auto
done

consts FUT ::'a



text{* The test specification:*}
test_spec "\<lbrakk>ids1 1 t; length t = length X; accept_ftp (rev t)\<rbrakk> \<Longrightarrow>
 (\<sigma>_0_ftp \<models> (os  \<leftarrow>  mbind t FTP_STP; return (os=X))) \<longrightarrow>
  FUT \<sigma>_0_ftp t = X"
apply(simp add: accept_ftp_def \<sigma>_0_ftp_def)
apply (rule impI)+
apply (unfold NB_ftp_def is_in_internet_def is_in_intranet_def)
apply (simp add:STSimps)
apply (gen_test_cases "FUT" split: HOL.split_if_asm simp:STSimps)
apply (simp_all add: POL1 POL2 POL3)
store_test_thm "ftp_test"

 
text{* We need to add all required lemmas to the simplifier set, such
that they can be used during test data generation. *}



lemmas ST_simps = Let_def valid_SE_def orelse_def unit_SE_def bind_SE_def
 in_subnet_def src_def dest_def IntegerPort.dest_port_def
 subnet_of_def id_def port_open_def is_init_def is_data_def
 is_port_request_def is_close_def p_accept_def content_def
 PolicyCombinators PortCombinators is_in_intranet_def
 is_in_internet_def intranet_def internet_def exI subnetOf_lemma
 subnetOf_lemma2 subnetOf_lemma3 subnetOf_lemma4 port_open_lemma
 ftp_policy_def allow_from_to_port_def

declare ST_simps[simp]
declare[[testgen_iterations=55000]]
declare[[testgen_SMT=false]]

gen_test_data "ftp_test"

thm ftp_test.test_data

declare ST_simps [simp del]

write_clocks "time_stats_FTP.tex"

end
