(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FWCompilation.thy --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Univerisite Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: FWCompilation.thy 9047 2010-11-17 18:23:54Z lukasbru $ *)

theory 
  Foo
imports
  FWTesting
begin




text{* Definition of the testing constraint, i.e. here we specify the
test scenario.  *}

definition all_positive :: "(IntegerPort_TCPUDP,'b) packet \<Rightarrow> bool" where
 "all_positive x = (id x > 0 \<and> src_port x > (0::port) \<and> dest_port x > (0::port))"




definition Adr1_1 :: "int set" where
 "Adr1_1 = {d. d > 0 \<and> d < 8}"

definition Adr1_2 :: "int set" where
 "Adr1_2 = {d. d > 7 \<and> d < 16}"

definition Adr1_3 :: "int set" where
 "Adr1_3 = {d. d > 15 \<and> d < 24}"

definition Adr1_4 :: "int set" where
 "Adr1_4 = {d. d > 23 \<and> d < 31}"

definition Adr1_5 :: "int set" where
 "Adr1_5 = {d. d = 32}"

definition Adr1_R :: "int set" where
 "Adr1_R = {d. d > 32 \<and> d < 256}"

definition Adr1 :: "int net" where
 "Adr1 = {Adr1_1,Adr1_2,Adr1_3,Adr1_4,Adr1_5}"


definition Adr2_1 :: "int set" where
 "Adr2_1 = {d. d = 101}"

definition Adr2_2 :: "int set" where
 "Adr2_2 = {d. d = 102}"

definition Adr2_3 :: "int set" where
 "Adr2_3 = {d. d > 115 \<and> d < 124}"

definition Adr2_4 :: "int set" where
 "Adr2_4 = {d. d > 123 \<and> d < 131}"

definition Adr2_5 :: "int set" where
 "Adr2_5 = {d. d = 132}"


definition Adr2_D :: "int set" where
 "Adr2_D = {d. d = 110}"

definition Adr2_R :: "int set" where
 "Adr2_R = {d. d > 132}"




definition Model1_1 :: "IntegerPort_TCPUDP set" where
 "Model1_1 = {(d,e). d \<in> Adr1_1}"

definition Model1_2 :: "IntegerPort_TCPUDP set" where
 "Model1_2 = {(d,e). d \<in> Adr1_2}"

definition Model1_3 :: "IntegerPort_TCPUDP set" where
 "Model1_3 = {(d,e). d \<in> Adr1_3}"

definition Model1_4 :: "IntegerPort_TCPUDP set" where
 "Model1_4 = {(d,e). d \<in> Adr1_4}"

definition Model1_5 :: "IntegerPort_TCPUDP set" where
 "Model1_5 = {(d,e). d \<in> Adr1_5}"

definition Model1_R :: "IntegerPort_TCPUDP set" where
 "Model1_R = {(d,e). d \<in> Adr1_R}"

definition Model1 :: "IntegerPort_TCPUDP net" where
 "Model1 = {Model1_1,Model1_2,Model1_3,Model1_4,Model1_5}"


definition Model2_1 :: "IntegerPort_TCPUDP set" where
 "Model2_1 = {(d,e). d \<in> Adr2_1}"

definition Model2_2 :: "IntegerPort_TCPUDP set" where
 "Model2_2 = {(d,e). d \<in> Adr2_2}"

definition Model2_3 :: "IntegerPort_TCPUDP set" where
 "Model2_3 = {(d,e). d \<in> Adr2_3}"

definition Model2_4 :: "IntegerPort_TCPUDP set" where
 "Model2_4 = {(d,e). d \<in> Adr2_4}"

definition Model2_5 :: "IntegerPort_TCPUDP set" where
 "Model2_5 = {(d,e). d \<in> Adr2_5}"


definition Model2_D :: "IntegerPort_TCPUDP set" where
 "Model2_D = {(d,e). d \<in> Adr2_D}"


definition Model2_R :: "IntegerPort_TCPUDP set" where
 "Model2_R = {(d,e). d \<in> Adr2_R}"


definition Model2 :: "IntegerPort_TCPUDP net" where
 "Model2 = {Model2_1,Model2_2,Model2_3,Model2_4,Model2_5,Model2_D}"



definition
  "not_in_same_net x  \<equiv>  (src x \<sqsubset> Model1 \<and> (dest x \<sqsubset> {Model2_R}))"



definition
 Filter :: "(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> unit" where
"Filter =     allow_from_to_ports_prot tcp {(8000::port)..8100} Model1 Model2 \<Oplus>
              allow_from_to_port_prot tcp   (21::port)            Model1 Model2 \<Oplus>
              allow_from_to_port_prot tcp   (80::port)            Model1 Model2 \<Oplus>
              allow_from_to_port_prot tcp  (443::port)            Model1 Model2 \<Oplus>
              allow_from_to_port_prot tcp   (554::port)              Model1 Model2 \<Oplus>
              allow_from_to_port_prot tcp   (25::port)              Model1 Model2 \<Oplus>
              allow_from_to_port_prot tcp   (22::port)             Model1 Model2 \<Oplus>
              deny_all"


definition NAT  ::"(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "NAT = pat2pool_int Adr1_1 Adr2_1"


definition NAT2  ::"(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "NAT2 = pat2pool_int Adr1_2 Adr2_2"

definition NAT3  ::"(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "NAT3 = nat2pool_int Adr1_3 Adr2_3"

definition NAT4  ::"(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "NAT4 = nat2pool_int Adr1_4 Adr2_4"

definition NAT5  ::"(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "NAT5 = nat2pool_int Adr1_5 Adr2_5"

definition NATID :: "(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "NATID = (\<lambda>p. Some (allow {p}))"

definition TestPolicy ::"(IntegerPort_TCPUDP,DummyContent) packet \<mapsto> (((IntegerPort_TCPUDP,DummyContent) packet) set)" where
 "TestPolicy = ((\<lambda> (x,y). x) o_f ((NAT \<Oplus> NAT2 \<Oplus>NAT3 \<Oplus> NAT4 \<Oplus> NAT5 \<Oplus> NATID) \<Otimes>\<^sub>2\<^sub>I Filter))"


lemmas policyLemmas = TestPolicy_def src_def dest_def 
                      src_port_def dest_port_def PolicyCombinators
                      PortCombinators in_subnet_def id_def
                      content_def IntegerPort_TCPUDP.src_port_def 
                      IntegerPort_TCPUDP.dest_port_def
                      dest_protocol_def allow_from_to_port_def
                      allow_from_to_ports_def NAT2_def 
                      NAT_def nat2pool_def prod_2_id_def prod_2_def
                      id_def content_def policy_range_comp_def
                      nat2pool_int_def NAT2_def NAT3_def NAT4_def NAT5_def
                      NATID_def

lemma either2[simp]: "(a \<noteq> tcp) = (a = udp)"
apply (case_tac a)
apply simp_all
done                      


lemma either3[simp]: "(a \<noteq> udp) = (a = tcp)"
apply (case_tac a)
apply simp_all
done                      


definition src_fix :: "(IntegerPort_TCPUDP,'b) packet \<Rightarrow> bool" where
 "src_fix x =  (fst (snd (src x)) = 1 \<and> (snd (snd (src x))) = udp)"

lemmas Models1 = Model1_1_def Model1_2_def Model1_3_def Model1_4_def Model1_5_def Model1_R_def Model1_def
lemmas Models2 = Model2_1_def Model2_2_def Model2_3_def Model2_4_def Model2_5_def Model2_D_def Model2_R_def Model2_def
lemmas Models = Models1 Models2


lemmas Adr1 = Adr1_1_def Adr1_2_def Adr1_3_def Adr1_4_def Adr1_5_def Adr1_R_def Adr1_def
lemmas Adr2 = Adr2_1_def Adr2_2_def Adr2_3_def Adr2_4_def Adr2_5_def Adr2_D_def Adr2_R_def 
lemmas Adr = Adr1 Adr2


test_spec "id x = 1 \<and> content x = DummyContent.data \<and> all_positive x \<and> not_in_same_net x \<and> src_fix x
      \<longrightarrow> FUT x = TestPolicy x"
apply (rule_tac x = x in spec)
apply (rule allI)
apply (simp add:   IntegerPort_TCPUDP.dest_protocol_def   
           policyLemmas Filter_def pat2pool_int_def Model1_def Model2_def Models Adr
 )
apply (gen_test_cases "FUT" simp add: Adr Models policyLemmas all_positive_def not_in_same_net_def src_fix_def src_protocol_def) 
apply (simp_all add: Models  Adr policyLemmas split_tupled_all Adr2)
store_test_thm "Scenario3"


declare [[testgen_iterations=50]]

ML {*
"Toplevel.debug := False"
*}

declare[[testgen_SMT=true]]

gen_test_data "Scenario3"

thm Scenario3.test_data
thm Scenario3.pos

export_fw_data "Scenario3c.data" Scenario3








end

datatype ('\<alpha>,'\<beta>) Combinators = 
  DenyAll
| DenyAllFromTo '\<alpha> '\<alpha> 
| AllowPortFromTo '\<alpha> '\<alpha> '\<beta> 
| Conc "(('\<alpha>,'\<beta>) Combinators)" "(('\<alpha>,'\<beta>) Combinators)" (infixr "\<oplus>" 80)

fun C :: "(IntegerPort net, port) Combinators \<Rightarrow> (IntegerPort,DummyContent) FWPolicy"
where
" C DenyAll = deny_all" 
|"C (DenyAllFromTo x y) = deny_all_from_to x y"
|"C (AllowPortFromTo x y p) = allow_from_to_port p x y"
|"C  (x \<oplus> y) =  C x ++ C y"


fun position :: "'\<alpha> \<Rightarrow> '\<alpha> list \<Rightarrow>  nat" where
  "position a [] = 0"
| "(position a (x#xs)) = (if a = x then 1 else (Suc (position a xs)))" 

fun srcNet where 
"srcNet (DenyAllFromTo x y) = x"
|"srcNet (AllowPortFromTo x y p) = x"

fun destNet where 
"destNet (DenyAllFromTo x y) = y"
|"destNet (AllowPortFromTo x y p) = y"

fun srcnets::"(IntegerPort net,port) Combinators \<Rightarrow> (IntegerPort net) list" where
 "srcnets DenyAll = [] " 
|"srcnets (DenyAllFromTo x y) = [x] " 
|"srcnets (AllowPortFromTo x y p) = [x] " 
|"(srcnets (x \<oplus> y)) = (srcnets x)@(srcnets y)"

fun destnets::"(IntegerPort net,port) Combinators \<Rightarrow> (IntegerPort net) list" where
 "destnets DenyAll = [] " 
|"destnets (DenyAllFromTo x y) = [y] " 
|"destnets (AllowPortFromTo x y p) = [y] "
|"(destnets (x \<oplus> y)) = (destnets x)@(destnets y)"

fun (sequential) net_list_aux where
 "net_list_aux [] = []"
|"net_list_aux (DenyAll#xs) = net_list_aux xs"
|"net_list_aux ((DenyAllFromTo x y)#xs) = x#y#(net_list_aux xs)"
|"net_list_aux ((AllowPortFromTo x y p)#xs) = x#y#(net_list_aux xs)"
|"net_list_aux ((x\<oplus>y)#xs) = (net_list_aux [x])@(net_list_aux [y])@(net_list_aux xs)"

fun net_list where "net_list p = remdups (net_list_aux p)"

definition bothNets where "bothNets x = (zip (srcnets x) (destnets x))"

fun (sequential) normBothNets where 
 "normBothNets ((a,b)#xs) = (if ((b,a) \<in> set xs) \<or> (a,b) \<in> set (xs) 
                            then (normBothNets xs)
                            else (a,b)#(normBothNets xs))"
|"normBothNets x = x" 

fun makeSets where 
 "makeSets ((a,b)#xs) = ({a,b}#(makeSets xs))"
|"makeSets [] = []"

fun bothNet where
 "bothNet DenyAll = {}"
|"bothNet (DenyAllFromTo a b) = {a,b}"
|"bothNet (AllowPortFromTo a b p) = {a,b}"

text{* $Nets\_List$ provides from a list of rules a list where the
entries are the appearing sets of source and destination network of each
rule. *}

definition Nets_List where "Nets_List x = makeSets (normBothNets (bothNets x))"

fun (sequential) first_srcNet where 
  "first_srcNet (x\<oplus>y) = first_srcNet x"
| "first_srcNet x = srcNet x"

fun (sequential) first_destNet where 
  "first_destNet (x\<oplus>y) = first_destNet x"
| "first_destNet x = destNet x"

fun (sequential) first_bothNet where
 "first_bothNet (x\<oplus>y) = first_bothNet x"
|"first_bothNet x = bothNet x"

fun (sequential) in_list where 
 "in_list DenyAll l = True"
|"in_list x l = (bothNet x \<in> set l)"

fun all_in_list where 
 "all_in_list [] l = True"
|"all_in_list (x#xs) l = (in_list x l \<and> all_in_list xs l)" 

fun (sequential) member where
 "member a (x\<oplus>xs) = ((member a x) \<or> (member a xs))"
|"member a x = (a = x)"

fun noneMT where 
 "noneMT (x#xs) = (dom (C x) \<noteq> {} \<and>  (noneMT xs))"
|"noneMT [] = True" 

fun notMTpolicy::"((IntegerPort net, port) Combinators) list \<Rightarrow> bool"  where
 "notMTpolicy (x#xs) = (if (dom (C x) = {}) then (notMTpolicy xs) else True)"
|"notMTpolicy [] = False"

fun sdnets where
  "sdnets DenyAll = {}"
| "sdnets (DenyAllFromTo a b) = {(a,b)}"
| "sdnets (AllowPortFromTo a b c) = {(a,b)}"
| "sdnets (a \<oplus> b) = sdnets a \<union> sdnets b"

definition packet_Nets  where "packet_Nets x a b = ((src x \<sqsubset> a \<and> dest x \<sqsubset> b) \<or>
                                                    (src x \<sqsubset> b \<and> dest x \<sqsubset> a))"

fun  matching_rule_rev where
"matching_rule_rev a (x#xs) = (if a \<in> dom (C x) then (Some x)
                               else (matching_rule_rev a xs))"
|"matching_rule_rev a [] = None"

text{* Provides the first matching rule of a policy given as a list of rules. *}
definition matching_rule where
"matching_rule a x =  (matching_rule_rev a (rev x))"

definition subnetsOfAdr where "subnetsOfAdr a = {x. a \<sqsubset> x}"

definition fst_set where "fst_set s = {a. \<exists> b. (a,b) \<in> s}"

definition snd_set where "snd_set s = {a. \<exists> b. (b,a) \<in> s}"

fun memberP where
 "memberP r (x#xs) = (member r x \<or> memberP r xs)"
|"memberP r [] = False"

fun firstList where 
 "firstList (x#xs) = (first_bothNet x)"
|"firstList [] = {}"

subsection{* Invariants *}

text{* If there is a DenyAll, it is at the first position *}
fun wellformed_policy1:: "((IntegerPort net, port) Combinators) list \<Rightarrow> bool" where 
  "wellformed_policy1 [] = True"
| "wellformed_policy1 (x#xs) = (DenyAll \<notin> (set xs))" 

text{* There is a DenyAll at the first position *}
fun wellformed_policy1_strong:: "((IntegerPort net, port) Combinators) list \<Rightarrow> bool"
where 
  "wellformed_policy1_strong [] = False"
| "wellformed_policy1_strong (x#xs) = (x=DenyAll \<and> (DenyAll \<notin> (set xs)))" 

text{* All rules appearing at the left of a DenyAllFromTo, have disjunct domains from it (except DenyAll) *}
fun (sequential) wellformed_policy2 where
  "wellformed_policy2 [] = True"
| "wellformed_policy2 (DenyAll#xs) = wellformed_policy2 xs"
| "wellformed_policy2 (x#xs) = ((\<forall> c a b. c = DenyAllFromTo a b \<and> c \<in> set xs \<longrightarrow>
                 Map.dom (C x) \<inter> Map.dom (C c) = {}) \<and> wellformed_policy2 xs)"

text{* An allow rule is disjunct with all rules appearing at the right of it. This invariant is not necessary as it is a consequence from others, but facilitates some proofs. *}
fun (sequential) wellformed_policy3::"((IntegerPort net, port) Combinators) list \<Rightarrow> bool" where
  "wellformed_policy3 [] = True"
| "wellformed_policy3 ((AllowPortFromTo a b p)#xs) = ((\<forall> r. r \<in> set xs \<longrightarrow>
      dom (C r) \<inter> dom (C (AllowPortFromTo a b p)) = {}) \<and> wellformed_policy3 xs)"
| "wellformed_policy3 (x#xs) = wellformed_policy3 xs"

text{* All two networks are either disjoint or equal. *}
definition netsDistinct where "netsDistinct a b = (\<not> (\<exists> x. x \<sqsubset> a \<and> x \<sqsubset> b))"

definition twoNetsDistinct where
  "twoNetsDistinct a b c d = (netsDistinct a c \<or> netsDistinct b d)"

definition allNetsDistinct where 
  "allNetsDistinct p = (\<forall> a b. (a \<noteq> b \<and> a \<in> set (net_list p) \<and>
                        b \<in> set (net_list p)) \<longrightarrow> netsDistinct a b)"

definition disjSD_2 where 
 "disjSD_2 x y = (\<forall> a b c d. ((a,b)\<in>sdnets x \<and>  (c,d) \<in>sdnets y \<longrightarrow>
      (twoNetsDistinct a b c d \<and> twoNetsDistinct a b d c)))"

text{* The policy is given as a list of single rules. *}
fun singleCombinators where 
"singleCombinators [] = True"
|"singleCombinators ((x\<oplus>y)#xs) = False"
|"singleCombinators (x#xs) = singleCombinators xs"

definition onlyTwoNets where
 "onlyTwoNets x = ((\<exists> a b. (sdnets x = {(a,b)})) \<or> (\<exists> a b. sdnets x = {(a,b),(b,a)}))" 

text{* Each entry of the list contains rules between two networks only. *}
fun OnlyTwoNets where 
 "OnlyTwoNets (DenyAll#xs) = OnlyTwoNets xs"
|"OnlyTwoNets (x#xs) = (onlyTwoNets x \<and> OnlyTwoNets xs)"
|"OnlyTwoNets [] = True"

fun noDenyAll where
 "noDenyAll (x#xs) = ((\<not> member DenyAll x) \<and> noDenyAll xs)"
|"noDenyAll [] = True"

fun noDenyAll1 where 
  "noDenyAll1 (DenyAll#xs) = noDenyAll xs"
| "noDenyAll1 xs = noDenyAll xs"

fun separated where
  "separated (x#xs) = ((\<forall> s. s \<in> set xs \<longrightarrow> disjSD_2 x s) \<and> separated xs)"
| "separated [] = True"

fun NetsCollected where 
  "NetsCollected (x#xs) = (((first_bothNet x \<noteq> firstList xs) \<longrightarrow>
          (\<forall>a\<in>set xs. first_bothNet x \<noteq> first_bothNet a)) \<and> NetsCollected (xs))"
| "NetsCollected [] = True"

fun NetsCollected2 where
 "NetsCollected2 (x#xs) = (xs = [] \<or> (first_bothNet x \<noteq> firstList xs \<and>
                           NetsCollected2 xs))"
|"NetsCollected2 [] = True"

subsection{* Transformations *}

text {*
The following two functions transform a policy into a list of single rules and vice-versa.
*}
fun policy2list::"(IntegerPort net, port) Combinators \<Rightarrow>
                 ((IntegerPort net, port) Combinators) list" where
 "policy2list (x \<oplus> y) = (concat [(policy2list x),(policy2list y)])"
|"policy2list x = [x]"

fun list2policy::"((IntegerPort net, port) Combinators) list \<Rightarrow>
                  ((IntegerPort net, port) Combinators)" where
 "list2policy (x#[]) = x"
| "list2policy (x#y) = x \<oplus> (list2policy y)"

text{* Remove all the rules appearing before a DenyAll. There are two alternative versions. *}

fun removeShadowRules1 where
  "removeShadowRules1 (x#xs) = (if (DenyAll \<in> set xs) 
                                then ((removeShadowRules1 xs))
                                else x#xs)"
| "removeShadowRules1 [] = []"

fun  removeShadowRules1_alternative_rev where
  "removeShadowRules1_alternative_rev [] = []"
| "removeShadowRules1_alternative_rev (DenyAll#xs) = [DenyAll]"
| "removeShadowRules1_alternative_rev [x] = [x]"
| "removeShadowRules1_alternative_rev (x#xs)=
                 x#(removeShadowRules1_alternative_rev xs)"

definition removeShadowRules1_alternative where
   "removeShadowRules1_alternative p =
                         rev (removeShadowRules1_alternative_rev (rev p))"

text{* Remove all the rules which allow a port, but are shadowed by a deny between these subnets *}

fun removeShadowRules2::  "((IntegerPort net, port) Combinators) list \<Rightarrow>
                          ((IntegerPort net, port) Combinators) list"
where
  "(removeShadowRules2 ((AllowPortFromTo x y p)#z)) = 
        (if (((DenyAllFromTo x y) \<in> set z)) 
         then ((removeShadowRules2 z))
         else (((AllowPortFromTo x y p)#(removeShadowRules2 z))))"
| "removeShadowRules2 (x#y) = x#(removeShadowRules2 y)"
| "removeShadowRules2 [] = []"

text{* Sorting a pocliy.  We first need to define an ordering on
rules.  This ordering depends on the $Nets\_List$ of a policy.  *}

fun smaller :: "(IntegerPort net, port) Combinators \<Rightarrow> 
                             (IntegerPort net, port) Combinators \<Rightarrow> 
                             ((IntegerPort net) set) list \<Rightarrow> bool" 
where
 "smaller DenyAll x l = True" 
| "smaller x DenyAll l = False"
| "smaller x y l = 
 ((x = y) \<or>         (if (bothNet x) = (bothNet y) then 
                       (case y of (DenyAllFromTo a b) \<Rightarrow> (x = DenyAllFromTo b a)  
                            | _ \<Rightarrow>  True)
                    else
                        (position (bothNet x) l <= position (bothNet y) l)))"


text{*
We use insertion sort for sorting a policy.
*}

fun insort where
  "insort a [] l = [a]"
| "insort a (x#xs) l = (if (smaller a x l) then a#x#xs else x#(insort a xs l))"

fun sort where
  "sort [] l = []"
| "sort (x#xs) l = insort x (sort xs l) l"

fun sorted  where
"sorted [] l \<longleftrightarrow> True" |
"sorted [x] l \<longleftrightarrow> True" |
"sorted (x#y#zs) l \<longleftrightarrow> smaller x y l \<and> sorted (y#zs) l"

text{*
@{text separate} works on a sorted policy: it joins the rules which talk about the traffic between the same two networks.
*}

fun separate where
 "separate (DenyAll#x) = DenyAll#(separate x)"
| "separate (x#y#z) = (if (first_bothNet x = first_bothNet y) 
                   then (separate ((x\<oplus>y)#z))
                   else (x#(separate(y#z))))"
|"separate x = x"                

text {*
Insert the DenyAllFromTo rules, such that traffic between two networks can be tested individually
*}

fun  insertDenies where
 "insertDenies (x#xs) = (case x of DenyAll \<Rightarrow> (DenyAll#(insertDenies xs))
                    | _  \<Rightarrow> (DenyAllFromTo (first_srcNet x) (first_destNet x) \<oplus> 
                       (DenyAllFromTo (first_destNet x) (first_srcNet x)) \<oplus> x)#
                                (insertDenies xs))"
| "insertDenies [] = []"

text{* Remove duplicate rules. This is especially necessary as
insertDenies might have inserted duplicate rules.

The second function is supposed to work on a list of policies. Only
rules which are duplicated within the same policy are removed.  *}


fun removeDuplicates where
  "removeDuplicates (x\<oplus>xs) = (if member x xs then (removeDuplicates xs)
                              else x\<oplus>(removeDuplicates xs))"
| "removeDuplicates x = x"

fun removeAllDuplicates where 
 "removeAllDuplicates (x#xs) = ((removeDuplicates (x))#(removeAllDuplicates xs))"
|"removeAllDuplicates x = x"

text{* Remove rules with an empty domain - they never match any packet. *}

fun removeShadowRules3::"((IntegerPort net, port) Combinators) list \<Rightarrow> ((IntegerPort net, port) Combinators) list"  where
 "removeShadowRules3 (x#xs) = (if (dom (C x)= {}) then (removeShadowRules3 xs)
                               else (x#(removeShadowRules3 xs)))"
|"removeShadowRules3 [] = []"

text {* Insert a DenyAll at the beginning of a policy. *}
fun insertDeny where 
 "insertDeny (DenyAll#xs) = DenyAll#xs"
|"insertDeny xs = DenyAll#xs"

text {* 
Now do everything:
*}

definition "sort' p l = sort l p"

definition 
  "normalize' p  = (removeAllDuplicates o insertDenies o separate o
                   (sort' (Nets_List p))  o removeShadowRules2 o remdups o
                   removeShadowRules3 o insertDeny o removeShadowRules1 o
                   policy2list) p"


definition normalize :: 
  "(IntegerPort net, port) Combinators \<Rightarrow> (IntegerPort net, port) Combinators list" 
where
   "normalize p = (removeAllDuplicates (insertDenies (separate (sort
                   (removeShadowRules2 (remdups (removeShadowRules3 (insertDeny
                  (removeShadowRules1 (policy2list p)))))) ((Nets_List p))))))"

definition
   "normalize_manual_order p l = removeAllDuplicates (insertDenies (separate
    (sort (removeShadowRules2 (remdups (removeShadowRules3 (insertDeny
    (removeShadowRules1 (policy2list p)))))) ((l)))))"

text{* Of course, normalize is equal to normalize', the latter looks nicer though. *}
lemma "normalize = normalize'"
by (rule ext, simp add: normalize_def normalize'_def sort'_def)

text{* The following definition helps in creating the test specification for the individual parts of a normalized policy. *}
definition makeFUT where 
   "makeFUT FUT p x n = 
     (packet_Nets x (fst (normBothNets (bothNets p)!n))  
                     (snd(normBothNets (bothNets p)!n)) \<longrightarrow> 
    FUT x = C ((normalize p)!Suc n) x)"


declare C.simps [simp del]

lemmas PLemmas = C.simps dom_def  PolicyCombinators.PolicyCombinators
                 PortCombinators.PortCombinators src_def dest_def in_subnet_def
                 IntegerPort.src_port_def IntegerPort.dest_port_def


declare dom_eq_empty_conv [simp del]

end
